from datetime import datetime

from authlib.flask.oauth2.sqla import (
    OAuth2ClientMixin,
    OAuth2TokenMixin,
    OIDCAuthorizationCodeMixin,
)
from flask_sqlalchemy import SQLAlchemy


db = SQLAlchemy()


class CRUDMixin:
    """Mixin that adds convenience methods for CRUD (create, read, update, delete) operations."""

    @classmethod
    def create(cls, **kwargs):
        """Create a new record and save it the database."""
        instance = cls(**kwargs)
        return instance.save()

    def update(self, commit=True, **kwargs):
        """Update specific fields of a record."""
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        if commit:
            self.save()
        return self

    def save(self, commit=True):
        """Save the record."""
        db.session.add(self)
        if commit:
            db.session.commit()
        return self

    def merge(self, commit=True):
        db.session.merge(self)
        if commit:
            db.session.commit()

    def delete(self, commit=True):
        """Remove the record from the database."""
        db.session.delete(self)
        return commit and db.session.commit()


class Model(CRUDMixin, db.Model):
    """Base model class that includes CRUD convenience methods."""

    __abstract__ = True


class Client(Model, OAuth2ClientMixin):
    # human readable name
    name = db.Column(db.String(40))
    # human readable description
    description = db.Column(db.String(400))

    client_id = db.Column(db.String(48), primary_key=True)
    client_secret = db.Column(db.String(120), unique=True, index=True, nullable=False)

    # public or confidential
    is_confidential = db.Column(db.Boolean)

    # TODO
    # approved = db.Column(db.Boolean, default=False)
    approved = True

    def check_response_type(self, response_type):
        return True

    def check_token_endpoint_auth_method(self, method):
        allowed = ["client_secret_post", "client_secret_basic"]
        if not self.is_confidential:
            allowed.append("none")
        return method in allowed

    def check_client_secret(self, secret):
        return self.client_secret == secret

    def check_grant_type(self, grant_type):
        return grant_type in ["authorization_code"]


class Grant(Model, OIDCAuthorizationCodeMixin):
    id = db.Column(db.Integer, primary_key=True)
    user = db.Column(db.String(40), nullable=False)

    client_id = db.Column(
        db.String(48), db.ForeignKey("client.client_id"), nullable=False
    )
    client = db.relationship("Client")
    expires_at = db.Column(db.DateTime)

    def is_expired(self):
        return self.expires_at < datetime.utcnow()

    def get_redirect_uri(self):
        return self.redirect_uri


class Token(Model, OAuth2TokenMixin):
    id = db.Column(db.Integer, primary_key=True)
    client_id = db.Column(
        db.String(48), db.ForeignKey("client.client_id"), nullable=False
    )
    client = db.relationship("Client")

    user = db.Column(db.String(40), nullable=False)
